/* link_archive.cpp */

#include "mld_internal.h"

static bool linker_separate_archive(finfo& input, std::vector<finfo>& ret, std::ostringstream& err) {

	size_t curPos = ARCHIVE_HEADER_LEN;

	ret.clear();
	
	bool is_first = true;

	for (;;) {

		// Is there room in input stream for another archiveMember (or are we at EOF)?

		if (curPos + sizeof(archiveMember) >= input.data.length()) {
			return true;
		}

		// Read header

		archiveMember am = { 0 };
		memcpy(&am, input.data.c_str() + curPos, sizeof(archiveMember));

		curPos += sizeof(archiveMember);

		// Is there room in input stream for declared data size?

		size_t dataSize = atoi(am.size);
		if (curPos + dataSize > input.data.length()) {
			err << "FATAL: Archive member declared size '" << dataSize << "', but only '" << (input.data.length() - curPos) << "' bytes remaining in file\n";
			return false; // corrupt
		}

		// Read data

		finfo nested;

		char buff[32] = { 0 };
		snprintf(buff, sizeof(buff), "%%0x%x", curPos);
		nested.filename = input.filename + buff;

		nested.data = input.data.substr(curPos, dataSize);
		curPos += dataSize;

		// Check whether this is a valid thing to parse

		bool is_metadata = (
			false
			|| (am.name[0] == '/' && am.name[1] == ' ') // IMAGE_ARCHIVE_LINKER_MEMBER
			|| (am.name[0] == '/' && am.name[1] == '/' && am.name[2] == ' ') // IMAGE_ARCHIVE_LINKER_LONGNAMES_MEMBER
		);

		// If so, stash it

		if (!is_metadata) {
			ret.push_back(nested);
		}

		// Round curPos up to nearest 2

		if (curPos % 2) {
			curPos++;
		}

	}

	// UNREACHED
}

static bool linker_file_is_archive(finfo& f) {
	return (f.data.substr(0, ARCHIVE_HEADER_LEN) == ARCHIVE_HEADER);
}

bool linker_separate_archives(linkOptions& lo) {

	bool should_recurse = false;

	// Expand any passed-in .a files into their constituent .obj files.

	for (int i = 0; i != lo.inputFiles.size() /* recheck every loop */; ++i) {

		if (linker_file_is_archive(lo.inputFiles[i])) {

			// It's an archive

			std::vector<finfo> separated;
			if (!linker_separate_archive(lo.inputFiles[i], separated, lo.lastError)) {
				lo.lastError << "FATAL: Failed to parse archive file '" << lo.inputFiles[i].filename << "'\n";
				return false;
			}

			// Check if we have nested archives
			for (int j = 0; j != separated.size(); ++j) {
				if (linker_file_is_archive(separated[j])) {
					should_recurse = true;
				}
			}

			// Append

			lo.inputFiles.insert(lo.inputFiles.end(), separated.begin(), separated.end());

			// Free the input archive, and shift back one index

			lo.inputFiles.erase(lo.inputFiles.begin() + i);
			i--;

		}

	}

	if (should_recurse) {
		return linker_separate_archives(lo);

	} else {
		return true;

	}
}
