/* link_input.cpp */

#include "linker.h"
#include <cstring> // strnlen

//
// single file processing
//

/*

coff object
- Symbol table declares what functions are present
- NO .reloc section, but there is IMAGE_RELOCATION (see numRelocations/ptrRelocations in COFF header)
- .obj-obj function linking is done as relocations (RVA / string table idx / relocationtype)
- .idata section describes dll imports (if any)

exe image
- No symbol table or string table
- Has .reloc section
- .idata section describes dll imports

*/

#define DIE(s) { \
	fputs(s "\n", stderr); \
	return 1; \
} 

int linker_process_file(bool verbose, finfo& loadedFile, coffInfo& ret) {

	uchar* o = loadedFile.data;

	// Attempt reading file as MZ/PE

	uchar* curPos = o;
	dosHeader* dh = (dosHeader*)curPos;
	curPos += sizeof(dosHeader);

	uint32 coffFileOffset = 0;

	if (dh->magic[0] == 'M' && dh->magic[1] == 'Z') {
		// Proceed as MZ/PE

		curPos = o + dh->pe_offset;

		// PE\0\0

		if (!(curPos[0] == 'P' && curPos[1] == 'E')) {
			DIE("Invalid PE magic");
		}

		curPos += 4; // PE\0\0
		coffFileOffset = curPos - o;

		if (verbose) {
			printf("Type:         Win32 MZ/PE (PE starts at 0x%X, COFF starts at 0x%X)\n",
				dh->pe_offset, coffFileOffset);
		}

	} else {
		// Not an MZ/PE file, assume raw COFF (.o/.obj) and reset stream
		curPos = o;

		if (verbose) {
			printf("Type:         Raw COFF\n");
		}
	}

	// Read COFF header

	coffHeader* ch = (coffHeader*)(curPos);
	curPos += sizeof(coffHeader);

	// Process COFF header

	if (verbose) {
		printf(
			"Machine type: %s (0x%x)\n",
			formatMachineType(ch->machine),
			ch->machine
			);

		printf(
			"Sections:     %u\n", ch->numSections);


		if (ch->timeStamp) {
			time_t rawtime = (time_t)ch->timeStamp;
			struct tm* ptm = gmtime(&rawtime);

			char tmBuff[256];
			strftime(tmBuff, 256, "%Y-%m-%d %H:%M:%S", ptm);

			printf(
				"Linked on:    %s UTC (%u)\n", tmBuff, ch->timeStamp);
		}

		printf("COFF symbols: ");
		if (ch->numSymbols) {
			printf("%d\n", ch->numSymbols);
		} else {
			printf("(not present)\n");
		}

		printf("Image header: ");
		if (ch->sizeOfOptionalHeader) {
			printf("present (%d bytes)\n", ch->sizeOfOptionalHeader);
		} else {
			printf("(not present)\n");
		}

	}

	// OPTIONAL HEADER


	ret.oh = NULL;
	if (ch->sizeOfOptionalHeader > 0) {


		optHeader* oh = (optHeader*)curPos;
		ret.oh = oh;
		curPos += sizeof(optHeader);

		if (oh->s.magic == 0x10B) {
			// OK - regular PE32 format

		} else if (oh->s.magic == 0x20B) {
			DIE("PE32+ detected, unsupported");

		} else {
			DIE("Invalid magic number in image header");
		}

		//.

		if (verbose) {

			printf("\nOPTIONAL HEADER\n");

			printf("Linker version:   %d.%d\n", oh->s.linkerVersionMajor, oh->s.linkerVersionMinor);

			printf("Total .text:      %d bytes\n", oh->s.sizeOfText);
			printf("Total .bss:       %d bytes\n", oh->s.sizeOfBSS);
			printf("Total .data:      %d bytes\n", oh->s.sizeOfData);

			printf("Entry point:      0x%08x\n", oh->s.addressOfEntryPoint);
			printf("Base of code:     0x%08x\n", oh->s.baseOfCode);

			printf("Image base:       0x%08x\n", oh->w.imageBase);
			printf("Section align:    %d bytes\n", oh->w.alignSections);
			printf("File align:       %d bytes\n", oh->w.alignFile);

			printf("Required OS:      %d.%d\n", oh->w.osVersionMajor, oh->w.osVersionMinor);
			printf("Image version:    %d.%d\n", oh->w.imageVersionMajor, oh->w.imageVersionMinor);
			printf("Subsystem ver:    %d.%d\n", oh->w.subsystemVersionMajor, oh->w.subsystemVersionMinor);

			printf("Subsystem:        %s (0x%x)\n",
				oh->w.subsystem == 0x3 ? "Console" : "other",
				oh->w.subsystem
				);

		}
	}

	/*
	This never happens
	if (ftell(f) - optHeaderStart != ch.sizeOfOptionalHeader) {
	uint32 skip = ch.sizeOfOptionalHeader - (ftell(f) - optHeaderStart);
	printf("Skipping over another %d byte(s)...\n", skip);
	fseek(f, skip, SEEK_CUR);
	}
	*/

	// SECTIONS

	std::map<std::string, nice_section> allSections;

	std::vector<sectionHeader*> rawSections;

	if (ch->numSections) {

		if (verbose) {
			printf("\nSECTIONS\n");
			printf(" Section virtualSize virtualAddr    filePos rawDataSize zeroPadding numRelocs\n");
		}

		for (int i = 0; i < ch->numSections; i++) {

			sectionHeader* sh = (sectionHeader*)curPos;
			rawSections.push_back(sh);
			curPos += sizeof(sectionHeader);

			nice_section ns;
			ns.raw = sh;
			ns.name = std::string(sh->name, strnlen(sh->name, 8));
			ns.data = std::string((char*)(o + sh->fpRawData), sh->sizeOfRawData);

			if (allSections.find(ns.name) != allSections.end()) {
				fprintf(stderr, "FATAL: Duplicate section '%s' in file\n", ns.name.c_str());
				return 1;
			}

			allSections[ns.name] = ns;
			ret.sectionNumbers.push_back(ns.name);

			if (verbose) {
				printf("%8.*s    %8d  0x%08x 0x%08x    %8d    %8d  %8d\n",
					8, sh->name,
					sh->virtualSize, sh->virtualAddress, sh->fpRawData, sh->sizeOfRawData,
					(sh->virtualSize > sh->sizeOfRawData
					? sh->virtualSize - sh->sizeOfRawData
					: 0
					),
					sh->numRelocations
					);
			}
		}
	}

	// [SECTION BODIES / RELOCATIONS / ETC GO HERE]

	// Relocations are reprocessed later once we have the string table, so
	//  this section is for verbose-output only

	if (verbose) {

		for (std::map<std::string, nice_section>::iterator i = allSections.begin(), e = allSections.end(); i != e; ++i) {
			if (i->second.raw->numRelocations) {

				printf("\nRELOCATIONS IN %s\n", i->first.c_str());
				printf("virtualAddress     relocType  symbolTableIdx\n");

				curPos = o + i->second.raw->fpRelocations;
				for (int j = 0; j < i->second.raw->numRelocations; j++) {
					relocation* reloc = (relocation*)curPos;
					curPos += SIZEOF_RELOCATION;

					printf(
						"    0x%08x  %12s  %d\n",
						reloc->virtualAddress,
						formatI386RelocationType(reloc->type),
						reloc->symbolTableIndex // index into symbol table
					);
				}
			}
		}

	} // (verbose)

	// [POST-SECTION BODIES]

	// COFF SYMBOLS

	curPos = o + ch->ptrSymbolTable;

	if (ch->numSymbols) {

		if (verbose) {
			printf("\nCOFF SYMBOLS\n");
			printf("      Name      Value Type  sectionNum storageClass\n");
		}

		int remainingAux = 0;
		coffSymbol* lastReal = NULL;

		ret.unresolvedExternals.clear();
		ret.providedExternals.clear();
		ret.providedInternals.clear();

		for (unsigned int i = 0; i < ch->numSymbols; i++) {
			coffSymbol* cs = (coffSymbol*)curPos;
			curPos += SIZEOF_COFFSYMBOL;

			if (!remainingAux) {
				// Normal record

				lastReal = cs;

				if (verbose) {

					if (cs->name.offset.zero == 0) {
						printf("s.t.%6d ", cs->name.offset.stringTableOffset); // string pointer relative to stringtablesizeptr not stringtableptr
					} else {
						printf("%10.*s ", 8, cs->name.shortName);
					}

					printf("0x%08x ", cs->value);

					if (cs->type == 0x20) {
						printf("func.");
					} else if (cs->type == 0x00) {
						printf("other");
					} else {
						printf("?    ");
					}

					printf("   %8hi %12s\n", cs->sectionNumber, formatStorageClass(cs->storageClass));

				}

				if (cs->numAuxSymbols) {
					remainingAux += cs->numAuxSymbols;
				}

				if (cs->storageClass == 2) { // STORAGE_CLASS_EXTERNAL
					if (cs->sectionNumber) {
						ret.providedExternals.push_back(cs);
					} else {
						ret.unresolvedExternals.push_back(cs);
					}
				} else {
					ret.providedInternals.push_back(cs);
				}

			} else {

				// We support some known types of auxilary record

				if (verbose) {

					if (lastReal->storageClass == 103) {
						// Auxilary record is 18-byte filename
						printf("@ Filename: %.*s\n", 18, (char*)cs);

					} else if (lastReal->storageClass == 2 && lastReal->type == 0x20 && lastReal->sectionNumber > 0) {
						printf("@ Function definition (function is >= %d bytes long)\n", cs->name.offset.stringTableOffset); // not really, but 4 bytes at a 4 byte offset

					} else if (lastReal->storageClass == 3 && allSections.find(std::string(lastReal->name.shortName, 8)) != allSections.end()) {
						printf("@ Section details\n");

					} else {
						printf("@ Auxilary record of unknown type\n");
					}

				}

				remainingAux--;
			}
		}

		// Next is the string table

		uint32* stringTableSizePtr = (uint32*)curPos;
		curPos += sizeof(uint32);
		uint32 stringTableSize = *stringTableSizePtr;

		uchar* stringTableStartPtr = curPos;

		if (stringTableSize > 4) {

			uint32 idx = 0;

			if (verbose) {
				printf("\nSTRING TABLE\n");
				printf("  idx  fptr Text\n");
			}

			for (;;) {
				if (verbose) {
					printf("%5d %5d %s\n",
						idx++,
						curPos - (uchar*)stringTableSizePtr,
						curPos
						);
				}

				ret.stringTableIndex.push_back((char*)curPos);

				curPos += strlen((char*)curPos) + 1; // trailing \0
				if (curPos >= o + loadedFile.flen || // EOF
					curPos >= stringTableStartPtr + stringTableSize // reached end of table
				) break;
			}

		}

		if (verbose) {

			if (ret.providedExternals.size()) {
				printf("\nPROVIDED EXTERNALS\n");
				for (unsigned int i = 0, e = ret.providedExternals.size(); i != e; ++i) {
					printf("%s\n", coffSymbol_getName(ret.providedExternals[i], stringTableSizePtr).c_str());
				}
			}

			if (ret.unresolvedExternals.size()) {
				printf("\nUNRESOLVED EXTERNALS\n");
				for (unsigned int i = 0, e = ret.unresolvedExternals.size(); i != e; ++i) {
					printf("%s\n", coffSymbol_getName(ret.unresolvedExternals[i], stringTableSizePtr).c_str());
				}
			}

		}

		ret.stringTableSizePtr = stringTableSizePtr;

	} else {
		ret.stringTableSizePtr = 0;
	}

	// Now that we are past the symbol table and string table, go back and
	//  process relocations into nice_relocation structs

	for (section_map_t::iterator ii = allSections.begin(), ee = allSections.end(); ii != ee; ++ii) {
		for (int iii = 0, eee = ii->second.raw->numRelocations; iii != eee; ++iii) {

			relocation* r = (relocation*)(o + ii->second.raw->fpRelocations + (iii * SIZEOF_RELOCATION));

			// look up symbol number
			coffSymbol* csym = (coffSymbol*)(o + ch->ptrSymbolTable + (r->symbolTableIndex * SIZEOF_COFFSYMBOL));

			// look up symbol name
			std::string symname = coffSymbol_getName(csym, ret.stringTableSizePtr);

			// produce nice_relocation and store along with section
			nice_relocation nr = { 0 };
			nr.symbol = symname;
			nr.type = r->type;
			nr.va = r->virtualAddress;
			ii->second.relocations.push_back(nr);

		}
	}

	ret.ch = ch;
	ret.allSections = allSections;

	return 0;

}
