/* mld/linker.cpp */

#include "linker.h"
#include <cstring> // memcpy
#include <cstdio> // snprintf

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

// Helpers

nice_symbolinfo processProvidedSymbol(coffSymbol* cs, int fileid, uint32* stringTableSizePtr, std::vector<std::string>* sectionNames) {
	nice_symbolinfo nss;
	nss.fileid        = fileid;
	nss.sectionOffset = cs->value;
	nss.symbol        = coffSymbol_getName(cs, stringTableSizePtr);
	if (cs->sectionNumber > 0) {
		nss.sectionName   = sectionNames->operator[]( cs->sectionNumber - 1); // coff section numbers are 1-based
	} else {
		// no section name present (this was an external reference)
	}

	return nss;
}

std::string getGlobalNameForLocalSymbol(int fileid, std::string symbolName) {
	char temp[1024] = { 0 };
	snprintf(temp, 1023, "@@%d@@%s", fileid, symbolName.c_str());
	return std::string(temp);
}

/*

- For each coff object:
-   check which functions are present
-   check which relocations refer to functions not present

 */

int linker_process(linkOptions& lo) {

	if (lo.baseAddress % lo.sectionAlign) {
		fputs("Fatal: base memory address must be aligned to section", stderr);
		return 1;
	}

	symbolmap_t
		allKnownSymbols,  // (all fields valid),
		allUnknownSymbols // (only useful field is 'symbol', fileid might be duplicated, value is zero)
	;

	sectiondata_t sectionData;

	vec_coffinfo_t fileinfo;

	sectioncharac_t sectionCharacteristics;

	// file_offsets stores the position in sectionData of each section from each source file
	offset_t file_offsets;

	// each file
	for (unsigned int i = 0, e = lo.inputFiles.size(); i != e; ++i) {
		coffInfo ci = { 0 };

		fprintf(stdout, "Processing %s...\n", lo.inputFiles[i].filename);

		int ret = linker_process_file(lo.verbose, lo.inputFiles[i], ci);

		if (ret) {
			fprintf(stderr, "Error loading file %s\n", lo.inputFiles[i].filename);
			return ret;
		}

		fileinfo.push_back(ci); // save for later

		// ========

		// Add all known symbols to allKnownSymbols (fail on duplicate)

		for (unsigned int ii = 0, ee = ci.providedExternals.size(); ii != ee; ++ii) {
			nice_symbolinfo ns = processProvidedSymbol(ci.providedExternals[ii], i, ci.stringTableSizePtr, &ci.sectionNumbers);

			if (allKnownSymbols.find(ns.symbol) != allKnownSymbols.end()) {
				fprintf(stderr, "FATAL: Duplicate symbol definition '%s' in files '%s' and '%s'\n",
					ns.symbol.c_str(),
					lo.inputFiles[allKnownSymbols[ns.symbol].fileid].filename,
					lo.inputFiles[i].filename
					);
				return 1;
			}

			allKnownSymbols[ns.symbol] = ns; // value copy
		}

		// Decorate all internal symbols and add them to allKnownSymbols ("no duplicates possible")
		for (unsigned int ii = 0, ee = ci.providedInternals.size(); ii != ee; ++ii) {
			nice_symbolinfo ns = processProvidedSymbol(ci.providedInternals[ii], i, ci.stringTableSizePtr, &ci.sectionNumbers);
			ns.symbol = getGlobalNameForLocalSymbol(i, ns.symbol); // override
			allKnownSymbols[ns.symbol] = ns; // value copy
		}

		// Add all unknown symbols to allUnknownSymbols (as long as they do not exist in allKnownSymbols). ignore on duplicate

		for (unsigned int ii = 0, ee = ci.unresolvedExternals.size(); ii != ee; ++ii) {
			nice_symbolinfo ns = processProvidedSymbol(ci.unresolvedExternals[ii], i, ci.stringTableSizePtr, &ci.sectionNumbers);

			if (allKnownSymbols.find(ns.symbol) != allKnownSymbols.end()) {
				// symbol known already
				printf("Unresolved symbol matched\n");
			}

			// ignore/overwrite on duplicate
			allUnknownSymbols[ns.symbol] = ns; // value copy
		}

		// =======

		// Identify all known sections

		for (section_map_t::iterator ii = ci.allSections.begin(), ee = ci.allSections.end(); ii != ee; ++ii) {

			if (lo.dropSections.find(ii->first) != lo.dropSections.end()) {
				continue; // don't include section
			}

			if (sectionData.find(ii->first) == sectionData.end()) {
				sectionData[ii->first] = std::string();
				file_offsets[ii->first] = mii_t();
			}
			file_offsets[ii->first][i] = sectionData[ii->first].length();
			sectionData[ii->first].append(ii->second.data);
			
			sectionCharacteristics[ii->first] = ii->second.raw->characteristics;
		}

	}

	// Filter allUnknownSymbols in case any of the new symbols matched
	unsigned int matchct = 0;
	for (symbolmap_t::iterator i = allUnknownSymbols.begin(), e = allUnknownSymbols.end(); i != e; ++i) {
		if (allKnownSymbols.find(i->first) != allKnownSymbols.end()) {
			matchct++;
		}
	}

	// fail if any symbols remain in allUnknownSymbols 
	if (matchct != allUnknownSymbols.size()) {
		fprintf(stderr, "FATAL: %d unresolved externals\n", allUnknownSymbols.size());
		return 1;

	} else {
		puts("All symbols match, linking...\n");

	}

	//
	// SECTIONS
	//

	
	// Retrieve section offsets
	
	sectionOffsetInfo info = sectionOffsetInfo(lo, sectionData);


	// Now that we have some idea of where all the data will reside in memory,
	//  it's possible to apply relocations.
	// n.b. must complete the above first to support inter-section relocations
	//  (two way - const data to function pointer; function const reference to
	//  global)

	// http://lists.cs.uiuc.edu/pipermail/llvm-commits/Week-of-Mon-20130624/179407.html 
	// http://www.iecc.com/linker/linker07.html


	for (int i = 0, e = fileinfo.size(); i != e; ++i) {
		for (section_map_t::iterator ii = fileinfo[i].allSections.begin(), ee = fileinfo[i].allSections.end(); ii != ee; ++ii) {

			if (ii->second.relocations.size() == 0) {
				continue; // no relocations
			}

			for (std::vector<nice_relocation>::iterator iii = ii->second.relocations.begin(), eee = ii->second.relocations.end(); iii != eee; ++iii) {
				nice_relocation& reloc = *iii;

				// look up symbol's location
				std::string reloc_symbol_name = reloc.symbol;
				if (allKnownSymbols.find(reloc_symbol_name) == allKnownSymbols.end()) {
					// Couldn't find symbol! Maybe it was a local symbol only.
					reloc_symbol_name = getGlobalNameForLocalSymbol(i, reloc.symbol); // local to source file?
					
					if (allKnownSymbols.find(reloc_symbol_name) == allKnownSymbols.end()) {
						fprintf(stderr, "Link: Couldn't resolve symbol '%s'!\n", reloc.symbol.c_str());
						return 1;
					}
				}

				nice_symbolinfo& nr = allKnownSymbols[reloc_symbol_name];

				uint32 source_section_va = lo.baseAddress + info.sectionMemPos[nr.sectionName] + file_offsets[nr.sectionName][nr.fileid];
				uint32 source_va = nr.sectionOffset + source_section_va;

				// find within-section position to write
				uint32 dest_section_foffset = file_offsets[ii->second.name /*section that contained reloc*/][i];
				uint32 dest_foffset = dest_section_foffset + reloc.va;

				// find VA position of write
				uint32 dest_section_moffset = lo.baseAddress + info.sectionMemPos[ii->second.name] + file_offsets[ii->second.name][i];
				uint32 dest_moffset = dest_section_moffset + reloc.va;

				// Analyse relocation type
				uint32 set_value = 0;

				if (reloc.type == RELOCATION_TYPE_DIR32) { // VA
					set_value = source_va;

				} else if (reloc.type == RELOCATION_TYPE_DIR32NB) { // RVA
					set_value = source_va - lo.baseAddress;

				} else if (reloc.type == RELOCATION_TYPE_REL32) { // PC-relative 

					// VA of relocation is relative to file
					uint32 pc = dest_moffset + 4;
					set_value = source_va - pc;

				} else {
					fprintf(stderr, "Unsupported relocation type 0x%x\n", reloc.type);
					return 1;
				}

				// Update section binary data
				memcpy(
					(char*)(sectionData[ii->second.name].c_str()) + dest_foffset,
					&set_value,
					4
				);

			}

		}
	}

	// Find RVA of entry point

	symbolmap_t::iterator entry_sym = allKnownSymbols.find(lo.entry_func);
	if (entry_sym == allKnownSymbols.end()) {
		fprintf(stderr, "Couldn\'t find entry point with symbol \"%s\"\n", lo.entry_func.c_str());
		return 1;
	}

	uint32 entryRva = (
		// RVA of section (i.e. not including base address offset)
		info.sectionMemPos[entry_sym->second.sectionName] + 

		// offset of source file within section
		file_offsets[entry_sym->second.sectionName][entry_sym->second.fileid] +

		// offset of symbol within source file
		entry_sym->second.sectionOffset
	);

	return linker_write_file(lo, sectionData, sectionCharacteristics, entryRva);
}
