/* mld/main.cpp */

#include <cstdio>
#include <cstring> // strcmp
#include <cstdlib> // atoi

#include "mld.h"

bool finfo_load(const char* fname, finfo& ret) {
	ret.filename = std::string(fname);

	FILE* f = fopen(fname, "rb");
	if (f == NULL) {
		return false;
	}

	fseek(f, 0, SEEK_END);
	unsigned long fsize = ftell(f);
	char* buff = new char[fsize];

	fseek(f, 0, SEEK_SET);
	if (fread(buff, 1, fsize, f) != fsize) {
		delete[] buff;
		return false;
	}

	fclose(f);

	ret.data.assign(buff, fsize);

	delete[] buff;

	return true;
}

int usage() {
	fputs(
		"Usage: mld [options]\n"
		"\n"
		"   --align-file {number}      Set section alignment in file (default 512)\n"
		"   --align-section {number}   Set section alignment in memory (default 4096)\n"
		"   --compress-rdata           Merge .rdata with .text to reduce filesize\n"
		"   --drop-section {name}      Do not include section in output\n"
		"   -e {symbol name}           Set entry point (default 'main')\n"
		"   -h                         Display this message\n"
		"   -i {filename}              Add input file to link\n"
		"   -mconsole                  Use CONSOLE subsystem (default)\n"
		"   -mwindows                  Use WINDOWS subsystem\n"
		"   -o {filename}              Set output file, - for stdout\n"
		"   -q                         Quiet (level 1); don't display linking errors\n"
		,
		stderr
	);
	return -1;
}

#define REQUIRE_FOLLOWING_ARG { if (i + 1 == argc) return usage(); }

int main(int argc, char** argv) {

	bool is_quiet = false;

	linkOptions lo;
	lo.entry_func = "main";

	for (int i = 1; i < argc; ++i) {

		if (!strcmp(argv[i], "--align-file")) {
			REQUIRE_FOLLOWING_ARG;
			lo.fileAlign = atoi(argv[++i]);

		} else if (!strcmp(argv[i], "--align-section")) {
			REQUIRE_FOLLOWING_ARG;
			lo.sectionAlign = atoi(argv[++i]);

		} else if (!strcmp(argv[i], "--compress-rdata")) {
			lo.compressRdata = true;

		} else if (!strcmp(argv[i], "--drop-section")) {
			REQUIRE_FOLLOWING_ARG;
			lo.dropSections.insert(std::string(argv[++i]));

		} else if (!strcmp(argv[i], "-e")) {
			REQUIRE_FOLLOWING_ARG;
			lo.entry_func = std::string(argv[++i]);

		} else if (!strcmp(argv[i], "-h")) {
			return usage();

		} else if (!strcmp(argv[i], "-q")) {
			is_quiet = true;

		} else if (!strcmp(argv[i], "-mwindows")) {
			lo.isWindowsSubsystem = true;

		} else if (!strcmp(argv[i], "-mconsole")) {
			lo.isWindowsSubsystem = false;

		} else if (!strcmp(argv[i], "-i")) {
			REQUIRE_FOLLOWING_ARG;
			finfo f;
			if (!finfo_load(argv[++i], f)) {
				fprintf(stderr, "Error loading file '%s'\n", argv[i]);
				return 1;
			}
			lo.inputFiles.push_back(f);

		} else if (!strcmp(argv[i], "-o")) {
			REQUIRE_FOLLOWING_ARG;
			lo.output = (strcmp(argv[++i], "-")) ? fopen(argv[i], "wb") : stdout;
			if (! lo.output) {
				fprintf(stderr, "Error opening output file '%s'\n", argv[i]);
				return 1;
			}

		} else {
			fprintf(stderr, "Unrecognised command '%s'\n", argv[i]);
			return usage();
		}

	}

	// APPLY

	bool linked = mld_link(lo);

	if (!is_quiet && lo.lastError.str().length() > 0) {
		fputs(lo.lastError.str().c_str(), stderr);
	}

	return linked ? 0 : 1;
}
