/* link_input.cpp */

#include "mld_internal.h"

//
// single file processing
//

/*

coff object
- Symbol table declares what functions are present
- NO .reloc section, but there is IMAGE_RELOCATION (see numRelocations/ptrRelocations in COFF header)
- .obj-obj function linking is done as relocations (RVA / string table idx / relocationtype)
- .idata section describes dll imports (if any)

exe image
- No symbol table or string table
- Has .reloc section
- .idata section describes dll imports

*/

#define DIE(s) { \
	serr << "FATAL: " s "\n"; \
	return false; \
} 

bool linker_process_file(finfo& loadedFile, coffInfo& ret, std::ostringstream& serr) {

	const char* inputfile_data = loadedFile.data.c_str();
	uchar* o = (uchar*)inputfile_data;

	// Attempt reading file as MZ/PE

	uchar* curPos = o;
	dosHeader* dh = (dosHeader*)curPos;
	curPos += sizeof(dosHeader);

	uint32 coffFileOffset = 0;

	if (dh->magic[0] == 'M' && dh->magic[1] == 'Z') {
		// Proceed as MZ/PE

		curPos = o + dh->pe_offset;

		// PE\0\0

		if (!(curPos[0] == 'P' && curPos[1] == 'E')) {
			DIE("Invalid PE magic");
		}

		curPos += 4; // PE\0\0
		coffFileOffset = curPos - o;

	} else {
		// Not an MZ/PE file, assume raw COFF (.o/.obj) and reset stream
		curPos = o;

	}

	// Read COFF header

	coffHeader* ch = (coffHeader*)(curPos);
	curPos += sizeof(coffHeader);

	// OPTIONAL HEADER
	
	ret.oh = NULL;
	if (ch->sizeOfOptionalHeader > 0) {

		optHeader* oh = (optHeader*)curPos;
		ret.oh = oh;
		curPos += sizeof(optHeader);

		if (oh->s.magic == 0x10B) {
			// OK - regular PE32 format

		} else if (oh->s.magic == 0x20B) {
			DIE("PE32+ detected, unsupported");

		} else {
			DIE("Invalid magic number in image header");
		}

		//.

	}

	// SECTIONS

	std::map<std::string, nice_section> allSections;

	std::vector<sectionHeader*> rawSections;

	if (ch->numSections) {

		for (int i = 0; i < ch->numSections; i++) {

			sectionHeader* sh = (sectionHeader*)curPos;
			rawSections.push_back(sh);
			curPos += sizeof(sectionHeader);

			nice_section ns;
			ns.raw = sh;
			ns.name = std::string(sh->name, strnlen(sh->name, 8)); // shortname. May be updated to longname later
			ns.data = std::string((char*)(o + sh->fpRawData), sh->sizeOfRawData);

			if (allSections.find(ns.name) != allSections.end()) {
				serr << "FATAL: Duplicate section '" << ns.name << "' in file\n";
				return false;
			}

			allSections[ns.name] = ns;
			ret.sectionNumbers.push_back(ns.name);

		}
	}

	// [POST-SECTION BODIES]

	// COFF SYMBOLS

	curPos = o + ch->ptrSymbolTable;

	if (ch->numSymbols) {

		int remainingAux = 0;
		coffSymbol* lastReal = NULL;

		ret.unresolvedExternals.clear();
		ret.providedExternals.clear();
		ret.providedInternals.clear();

		for (unsigned int i = 0; i < ch->numSymbols; i++) {
			coffSymbol* cs = (coffSymbol*)curPos;
			curPos += SIZEOF_COFFSYMBOL;

			if (!remainingAux) {
				// Normal record

				lastReal = cs;

				if (cs->numAuxSymbols) {
					remainingAux += cs->numAuxSymbols;
				}

				if (cs->storageClass == 2) { // STORAGE_CLASS_EXTERNAL
					if (cs->sectionNumber) {
						ret.providedExternals.push_back(cs);
					} else {
						ret.unresolvedExternals.push_back(cs);
					}
				} else {
					ret.providedInternals.push_back(cs);
				}

			} else {

				// We support some known types of auxilary record

				remainingAux--;
			}
		}

		// Next is the string table

		uint32* stringTableSizePtr = (uint32*)curPos;
		curPos += sizeof(uint32);
		uint32 stringTableSize = *stringTableSizePtr;

		uchar* stringTableStartPtr = curPos;

		if (stringTableSize > 4) {

			uint32 idx = 0;

			for (;;) {
				ret.stringTableIndex.push_back((char*)curPos);

				curPos += strlen((char*)curPos) + 1; // trailing \0
				if (curPos >= o + loadedFile.data.size() || // EOF
					curPos >= stringTableStartPtr + stringTableSize // reached end of table
				) break;
			}

			// Fix up any sections that used longnames {{ 

			section_map_t replacement;

			for (section_map_t::iterator ii = allSections.begin(), ee = allSections.end(); ii != ee; ++ii) {
				if (ii->first[0] == '/') {
					uint32 section_longname_stringtable_offset = atoi(ii->first.c_str() + 1);
					std::string replacementName = (char*)(stringTableStartPtr - sizeof(uint32) + section_longname_stringtable_offset);

					replacement[replacementName] = ii->second;

				} else {
					replacement[ii->first] = ii->second;
				}
			}

			allSections = replacement;

			// }} 

		}

		ret.stringTableSizePtr = stringTableSizePtr;

	} else {
		ret.stringTableSizePtr = 0;
	}

	// Now that we are past the symbol table and string table, go back and
	//  process relocations into nice_relocation structs

	for (section_map_t::iterator ii = allSections.begin(), ee = allSections.end(); ii != ee; ++ii) {
		for (int iii = 0, eee = ii->second.raw->numRelocations; iii != eee; ++iii) {

			relocation* r = (relocation*)(o + ii->second.raw->fpRelocations + (iii * SIZEOF_RELOCATION));

			// look up symbol number
			coffSymbol* csym = (coffSymbol*)(o + ch->ptrSymbolTable + (r->symbolTableIndex * SIZEOF_COFFSYMBOL));

			// look up symbol name
			std::string symname = coffSymbol_getName(csym, ret.stringTableSizePtr);

			// produce nice_relocation and store along with section
			nice_relocation nr = { 0 };
			nr.symbol = symname;
			nr.type = r->type;
			nr.va = r->virtualAddress;
			ii->second.relocations.push_back(nr);

		}
	}

	ret.ch = ch;
	ret.allSections = allSections;

	return true;
}
