/* mld/linker.cpp */

#include "linker.h"
#include <cstring> // memcpy
#include <cstdio> // snprintf
#include <algorithm> // std::sort

#ifdef _MSC_VER
#define snprintf _snprintf
#endif

// Helpers

std::string coffSymbol_getName(coffSymbol* cs, uint32* stringTableSizePtr) {
	if (cs->name.offset.zero == 0) {
		char* str = (char*)(stringTableSizePtr)+cs->name.offset.stringTableOffset;
		return str;

	} else {
		return std::string(cs->name.shortName, strnlen(cs->name.shortName, 8));
	}
}

nice_symbolinfo processProvidedSymbol(coffSymbol* cs, int fileid, uint32* stringTableSizePtr, std::vector<std::string>* sectionNames) {
	nice_symbolinfo nss;
	nss.fileid        = fileid;
	nss.sectionOffset = cs->value;
	nss.symbol        = coffSymbol_getName(cs, stringTableSizePtr);
	if (cs->sectionNumber > 0) {
		nss.sectionName   = sectionNames->operator[]( cs->sectionNumber - 1); // coff section numbers are 1-based
	} else {
		// no section name present (this was an external reference)
	}

	return nss;
}

std::string getGlobalNameForLocalSymbol(int fileid, std::string symbolName) {
	char temp[1024] = { 0 };
	snprintf(temp, 1023, "@@%d@@%s", fileid, symbolName.c_str());
	return std::string(temp);
}

std::string section_getBaseName(std::string in) {

	// Filter .base$XXXX
	size_t cpos = in.find_first_of('$', 0);
	if (cpos != std::string::npos) {
		return in.substr(0, cpos); 
	}

	// No filters remaining
	return in;
}

void walkSymbols(int fileid, std::map<int, bool> &necessary_files, vec_coffinfo_t &fileinfo, symbolmap_t &allKnownSymbols) {

	// Visited self
	necessary_files[fileid] = true;

	// For each symbol required for this file,
	for (int i = 0, e = fileinfo[fileid].unresolvedExternals.size(); i != e; ++i) {

		nice_symbolinfo ns = processProvidedSymbol(
			fileinfo[fileid].unresolvedExternals[i], fileid, fileinfo[fileid].stringTableSizePtr, &fileinfo[fileid].sectionNumbers
		);

		// determine which other file provided it
		int providing_fileid = allKnownSymbols[ns.symbol].fileid;

		// check if file is not yet visited
		if (necessary_files.find(providing_fileid) == necessary_files.end()) {
			walkSymbols(providing_fileid, necessary_files, fileinfo, allKnownSymbols);
		}

	}
	
}

/*

- For each coff object:
-   check which functions are present
-   check which relocations refer to functions not present

 */

int linker_process(linkOptions& lo) {

	if (lo.baseAddress % lo.sectionAlign) {
		fputs("Fatal: base memory address must be aligned to section", stderr);
		return 1;
	}

	symbolmap_t
		allKnownSymbols,  // (all fields valid),
		allUnknownSymbols // (only useful field is 'symbol', fileid might be duplicated, value is zero)
	;

	sectiondata_t sectionData;

	vec_coffinfo_t fileinfo;

	sectioncharac_t sectionCharacteristics;

	std::vector<sectionIntentToCopy> sectionOrdering;

	// file_offsets stores the position in sectionData of each section from each source file
	offset_t file_offsets;

	// each file
	for (unsigned int i = 0, e = lo.inputFiles.size(); i != e; ++i) {
		coffInfo ci = { 0 };

		// fprintf(stdout, "Processing '%s'...\n", lo.inputFiles[i].filename.c_str());

		int ret = linker_process_file(lo.inputFiles[i], ci);

		if (ret) {
			fprintf(stderr, "Error loading file '%s'\n", lo.inputFiles[i].filename.c_str());
			return ret;
		}

		fileinfo.push_back(ci); // save for later

		// ========

		// Add all known symbols to allKnownSymbols (fail on duplicate)

		for (unsigned int ii = 0, ee = ci.providedExternals.size(); ii != ee; ++ii) {
			nice_symbolinfo ns = processProvidedSymbol(ci.providedExternals[ii], i, ci.stringTableSizePtr, &ci.sectionNumbers);

			if (allKnownSymbols.find(ns.symbol) != allKnownSymbols.end()) {
				fprintf(stderr, "FATAL: Duplicate symbol definition '%s' in files '%s' and '%s'\n",
					ns.symbol.c_str(),
					lo.inputFiles[allKnownSymbols[ns.symbol].fileid].filename.c_str(),
					lo.inputFiles[i].filename.c_str()
					);
				return 1;
			}

			allKnownSymbols[ns.symbol] = ns; // value copy
		}

		// Decorate all internal symbols and add them to allKnownSymbols ("no duplicates possible")
		for (unsigned int ii = 0, ee = ci.providedInternals.size(); ii != ee; ++ii) {
			nice_symbolinfo ns = processProvidedSymbol(ci.providedInternals[ii], i, ci.stringTableSizePtr, &ci.sectionNumbers);
			ns.symbol = getGlobalNameForLocalSymbol(i, ns.symbol); // override
			allKnownSymbols[ns.symbol] = ns; // value copy
		}

		// Add all unknown symbols to allUnknownSymbols (as long as they do not exist in allKnownSymbols). ignore on duplicate

		for (unsigned int ii = 0, ee = ci.unresolvedExternals.size(); ii != ee; ++ii) {
			nice_symbolinfo ns = processProvidedSymbol(ci.unresolvedExternals[ii], i, ci.stringTableSizePtr, &ci.sectionNumbers);

			if (allKnownSymbols.find(ns.symbol) != allKnownSymbols.end()) {
				// symbol known already
				// printf("Unresolved symbol matched\n");
			}

			// ignore/overwrite on duplicate
			allUnknownSymbols[ns.symbol] = ns; // value copy
		}

		// =======

		// Identify all known sections

		for (section_map_t::iterator ii = ci.allSections.begin(), ee = ci.allSections.end(); ii != ee; ++ii) {

			std::string strippedSectionName = section_getBaseName(ii->first);

			if (
				lo.dropSections.find(ii->first) != lo.dropSections.end() ||
				lo.dropSections.find(strippedSectionName) != lo.dropSections.end()
			) {
				continue; // don't include section
			}

			// 

			sectionIntentToCopy element;
			element.sourceSectionName = ii->first;
			element.targetSectionName = strippedSectionName;
			element.source_fileno = i;

			sectionOrdering.push_back(element);

		}

	}

	// Filter allUnknownSymbols in case any of the new symbols matched
	unsigned int matchct = 0;
	for (symbolmap_t::iterator i = allUnknownSymbols.begin(), e = allUnknownSymbols.end(); i != e; ++i) {
		if (allKnownSymbols.find(i->first) != allKnownSymbols.end()) {
			matchct++;
		}
	}

	// fail if any symbols remain in allUnknownSymbols 
	if (matchct != allUnknownSymbols.size()) {
		fprintf(stderr, "FATAL: %d unresolved externals\n", allUnknownSymbols.size());
		return 1;

	} else {
		// puts("All symbols match, linking...\n");

	}

	// fail if entry point not found
	symbolmap_t::iterator entry_sym = allKnownSymbols.find(lo.entry_func);
	if (entry_sym == allKnownSymbols.end()) {
		fprintf(stderr, "Couldn\'t find entry point with symbol \"%s\"\n", lo.entry_func.c_str());
		return 1;
	}

	// Identify which files are actually necessary

	std::map<int, bool> necessary_files;
	walkSymbols(entry_sym->second.fileid, necessary_files, fileinfo, allKnownSymbols);

	// Strip out sections from unnecessary files
	{
		std::vector<sectionIntentToCopy> replacement;
		for (size_t i = 0, e = sectionOrdering.size(); i != e; ++i) {
			if (necessary_files.find(sectionOrdering[i].source_fileno) == necessary_files.end()) {
				// not found, not necessary
			} else {
				// found, neccessary
				replacement.push_back(sectionOrdering[i]);
			}
		}

		sectionOrdering = replacement;
	}

	// Sort sections

	std::sort(sectionOrdering.begin(), sectionOrdering.end());

	// Special handling for cygwin/mingw handling of .idata sections
	//  (necessary if we want to use their import libraries)

	//.idata BLOCK(__section_alignment__) :
	//{
	//	/* This cannot currently be handled with grouped sections.
	//	See pe.em:sort_sections.  */
	//	SORT(*)(.idata$2)
	//		SORT(*)(.idata$3)
	//		/* These zeroes mark the end of the import list.  */
	//		LONG(0); LONG(0); LONG(0); LONG(0); LONG(0);
	//	SORT(*)(.idata$4)
	//		__IAT_start__ = .;
	//	SORT(*)(.idata$5)
	//		__IAT_end__ = .;
	//	SORT(*)(.idata$6)
	//		SORT(*)(.idata$7)
	//}

	// Build sectionData in memory

	uint32 idata_state_tracker = 0;
	uint32 iat_start_offset = 0;
	uint32 iat_end_offset = 0;
	
	for (size_t i = 0, e = sectionOrdering.size(); i < e; ++i) {

		sectionIntentToCopy &element = sectionOrdering[i];

		// Special handling for cygwin/mingw .idata sections (necessary if we want to use their import libraries)
		// See /mld/doc/cygwin-default-linker-script.lds or `gcc -Wl,--verbose` on any test C file
		// {{{
		switch (idata_state_tracker) {

			case 0: {
				if (element.sourceSectionName == ".idata$4") {
					sectionData[".idata"].append(
						"\x00\x00\x00\x00" "\x00\x00\x00\x00" "\x00\x00\x00\x00""\x00\x00\x00\x00" "\x00\x00\x00\x00",
						5 * 4 * sizeof(char) // five
					);
					idata_state_tracker = 1;
				}
			} break;

			case 1: {
				if (element.sourceSectionName == ".idata$5") {
					iat_start_offset = sectionData[".idata"].length();
					idata_state_tracker = 2;
				}
			} break;

			case 2: {
				if (element.sourceSectionName == ".idata$6") {
					iat_end_offset = sectionData[".idata"].length();
					idata_state_tracker = 3; // end
				}
			} break;

		}
		// }}}

		// 

		if (sectionData.find(element.targetSectionName) == sectionData.end()) {
			sectionData[element.targetSectionName] = std::string();
		}

		if (file_offsets.find(element.sourceSectionName) == file_offsets.end()) {
			file_offsets[element.sourceSectionName] = mii_t();
		}

		file_offsets[element.sourceSectionName][element.source_fileno] = sectionData[element.targetSectionName].length(); // save start position

		sectionData[element.targetSectionName].append(fileinfo[element.source_fileno].allSections[element.sourceSectionName].data); // append binary data

		sectionCharacteristics[element.targetSectionName] = fileinfo[element.source_fileno].allSections[element.sourceSectionName].raw->characteristics;
	}

	// Strip sections with zero virtual-size

	{
		sectiondata_t::iterator i = sectionData.begin();
		for (i = sectionData.begin(); i != sectionData.end();) {
			if (!i->second.size()) {
				i = sectionData.erase(i);
			} else {
				++i;
			}
		}
	}

	//
	// SECTIONS
	//

	
	// Retrieve section offsets
	
	sectionOffsetInfo info = sectionOffsetInfo(lo, sectionData);


	// Now that we have some idea of where all the data will reside in memory,
	//  it's possible to apply relocations.
	// n.b. must complete the above first to support inter-section relocations
	//  (two way - const data to function pointer; function const reference to
	//  global)

	// http://lists.cs.uiuc.edu/pipermail/llvm-commits/Week-of-Mon-20130624/179407.html 
	// http://www.iecc.com/linker/linker07.html


	for (int i = 0, e = fileinfo.size(); i != e; ++i) {

		if (necessary_files.find(i) == necessary_files.end()) {
			// we skipped this file, so don't apply relocations from it
			continue;
		}

		for (section_map_t::iterator ii = fileinfo[i].allSections.begin(), ee = fileinfo[i].allSections.end(); ii != ee; ++ii) {

			if (ii->second.relocations.size() == 0) {
				continue; // no relocations
			}

			for (std::vector<nice_relocation>::iterator iii = ii->second.relocations.begin(), eee = ii->second.relocations.end(); iii != eee; ++iii) {
				nice_relocation& reloc = *iii;

				// look up symbol's location
				std::string reloc_symbol_name = reloc.symbol;
				if (allKnownSymbols.find(reloc_symbol_name) == allKnownSymbols.end()) {
					// Couldn't find symbol! Maybe it was a local symbol only.
					reloc_symbol_name = getGlobalNameForLocalSymbol(i, reloc.symbol); // local to source file?
					
					if (allKnownSymbols.find(reloc_symbol_name) == allKnownSymbols.end()) {
						fprintf(stderr, "Link: Couldn't resolve symbol '%s'!\n", reloc.symbol.c_str());
						return 1;
					}
				}

				nice_symbolinfo& nr = allKnownSymbols[reloc_symbol_name];

				uint32 source_section_va = lo.baseAddress + info.sectionMemPos[section_getBaseName(nr.sectionName)] + file_offsets[nr.sectionName][nr.fileid];
				uint32 source_va = nr.sectionOffset + source_section_va;

				// find within-section position to write
				uint32 dest_section_foffset = file_offsets[ii->second.name /*section that contained reloc*/][i];
				uint32 dest_foffset = dest_section_foffset + reloc.va;

				// find VA position of write
				uint32 dest_section_moffset = lo.baseAddress + info.sectionMemPos[section_getBaseName(ii->second.name)] + file_offsets[ii->second.name][i];
				uint32 dest_moffset = dest_section_moffset + reloc.va;

				// Analyse relocation type
				uint32 set_value = 0;

				if (reloc.type == RELOCATION_TYPE_DIR32) { // VA
					set_value = source_va;

				} else if (reloc.type == RELOCATION_TYPE_DIR32NB) { // RVA
					set_value = source_va - lo.baseAddress;

				} else if (reloc.type == RELOCATION_TYPE_REL32) { // PC-relative 

					// VA of relocation is relative to file
					uint32 pc = dest_moffset + 4;
					set_value = source_va - pc;

				} else {
					fprintf(stderr, "Unsupported relocation type 0x%x\n", reloc.type);
					return 1;
				}

				// Update section binary data
				memcpy(
					(char*)(sectionData[section_getBaseName(ii->second.name)].c_str()) + dest_foffset,
					&set_value,
					sizeof(uint32) // 4 
				);

			}

		}
	}

	// Find RVA of entry point

	uint32 entryRva = (
		// RVA of section (i.e. not including base address offset)
		info.sectionMemPos[entry_sym->second.sectionName] + 

		// offset of source file within section
		file_offsets[entry_sym->second.sectionName][entry_sym->second.fileid] +

		// offset of symbol within source file
		entry_sym->second.sectionOffset
	);


	return linker_write_file(lo, sectionData, sectionCharacteristics, entryRva, iat_start_offset, iat_end_offset);
}
